#!/usr/bin/python3
import gpiozero

debounce = 3e-3 # 3 ms

class Incremental_Encoder:
    """Read out an incremental encoder from GPIO pins

    Instance Variables
    ------------------
    
    The following instance variables may be set to handle events. At least one
    of them should be set to make the device functional at all.

    on_turn : function(direction)
        Function to call when encoder is turned.
        'direction' is either 1 or -1 indicating a cw or ccw rotation.

    on_turn_cw: function()
    on_turn_ccw: function()
        Function to call when encoder is turned in the named direction.

    Decoding
    --------

    The encoder has two switches to ground. They are connected to pins with
    pull-up and generate the following signal curves:

                  ,---,   ,---,   ,---,
                  |off|on |   |   |   |
        Pin A  ---'   '---'   '---'   '---
               -,   ,---,   ,---,   ,---,
                |   |off|on |   |   |   |
        Pin B   '---'   '---'   '---'   '-
        Detent      ^       ^       ^      -----> cw rotation

    While switch A is closed (clock) switch B indicates direction.

    Reference: ALPS 402160, Reichelt STEC12E07
    """
    def __init__(self, pin_A, pin_B, reverse=False):
        """
        pin_A, pin_B : int or string
            GPIO pins for switch A and B

        reverse : boolean
            If True, the direction of rotation is reversed.
        """
        self.reverse = reverse
        self.A = gpiozero.Button(pin_A, pull_up=True, bounce_time=debounce)
        self.B = gpiozero.Button(pin_B, pull_up=True, bounce_time=debounce)
        self.A.when_pressed = self._press_A
        self.B.when_pressed = self._press_B
        self.B.when_released = self._release_B

        self._state_B = 'on' if self.B.value else 'off'

        self.on_turn = None
        self.on_turn_cw = None
        self.on_turn_ccw = None

    # evaluate direction of rotation
    #
    # Edge-triggered callback function are called with a reasonable lag. Also
    # the read-out of a pin state has a lag. So it is impossible to ask for
    # the state of pin B when _press_A() is called. Thus, the state of pin B
    # has to be remembered in an instance variable.

    def _press_A(self):
        if self._state_B == 'off':
            self._ev_rotation(cw=True)
        else:
            self._ev_rotation(cw=False)

    def _press_B(self):
        self._state_B = 'on'

    def _release_B(self):
        self._state_B = 'off'
        
    # call external event handler
    
    def _ev_rotation(self, cw):
        if self.reverse:
            cw = not cw
        if cw:
            direction = 1
            if self.on_turn_cw: self.on_turn_cw()
        else:
            direction = -1
            if self.on_turn_ccw: self.on_turn_ccw()
        if self.on_turn: self.on_turn(direction)
            
# ----------------------------------------------------------------------------
# Test
# ----------------------------------------------------------------------------

if __name__ == "__main__":
    from signal import pause

    def _test_ev_handler(direction):
        if direction == 1:
            print('right')
        elif direction == -1:
            print('left')
        else:
            print('hugh?')

    my_knob = Incremental_Encoder(6, 5)
    my_knob.on_turn = _test_ev_handler

    pause()

# vim: set et:
